@@ -1,5 +1,5 @@
 /*  Lzip - A data compressor based on the LZMA algorithm
-    Copyright (C) 2008 Antonio Diaz Diaz.
+    Copyright (C) 2008, 2009, 2010 Antonio Diaz Diaz.
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -17,32 +17,31 @@
 
 class Input_buffer
   {
-  const int buffer_size;
+  enum { buffer_size = 65536 };
   uint8_t * const buffer;
   int pos;
-  int size;		// when reached, a new block must be read
-  const int _ides;
-  bool _finished;
+  int stream_pos;		// when reached, a new block must be read
+  const int ides_;		// input file descriptor
+  bool at_stream_end;
 
   bool read_block();
 
 public:
   Input_buffer( const int ides )
     :
-    buffer_size( 65536 ),
     buffer( new uint8_t[buffer_size] ),
     pos( 0 ),
-    size( 0 ),
-    _ides( ides ),
-    _finished( false ) {}
+    stream_pos( 0 ),
+    ides_( ides ),
+    at_stream_end( false ) {}
 
   ~Input_buffer() { delete[] buffer; }
 
-  bool finished() const throw() { return _finished; }
+  bool finished() const throw() { return at_stream_end && pos >= stream_pos; }
 
-  uint8_t read_byte()
+  uint8_t get_byte()
     {
-    if( pos >= size && !read_block() ) return 0;
+    if( pos >= stream_pos && !read_block() ) return 0;
     return buffer[pos++];
     }
   };
@@ -50,7 +49,7 @@ public:
 
 class Range_decoder
   {
-  long long file_pos;
+  mutable long long member_pos;
   uint32_t code;
   uint32_t range;
   Input_buffer & ibuf;
@@ -58,61 +57,79 @@ class Range_decoder
 public:
   Range_decoder( const int header_size, Input_buffer & buf )
     :
-    file_pos( header_size ),
-    code( 0 ), range( 0xFFFFFFFF ),
+    member_pos( header_size ),
+    code( 0 ),
+    range( 0xFFFFFFFF ),
     ibuf( buf )
-    { for( int i = 0; i < 5; ++i ) code = (code << 8) | read_byte(); }
+    { for( int i = 0; i < 5; ++i ) code = (code << 8) | get_byte(); }
 
-  uint8_t read_byte()
+  bool code_is_zero() const throw() { return ( code == 0 ); }
+  bool finished() const throw() { return ibuf.finished(); }
+  long long member_position() const throw() { return member_pos; }
+
+  uint8_t get_byte() const
     {
-    ++file_pos;
-    return ibuf.read_byte();
+    ++member_pos;
+    return ibuf.get_byte();
     }
 
-  long long file_position() const throw() { return file_pos; }
-  bool finished() const throw() { return ibuf.finished(); }
+  void reload() throw()
+    {
+    code = 0;
+    range = 0xFFFFFFFF;
+    for( int i = 0; i < 5; ++i ) code = (code << 8) | get_byte();
+    }
 
-  int decode( const int num_bits = 1 )
+  void normalize()
+    {
+    if( range <= 0x00FFFFFF )
+      { range <<= 8; code = (code << 8) | get_byte(); }
+    }
+
+  int decode( const int num_bits )
     {
     int symbol = 0;
-    for( int i = num_bits - 1; i >= 0; --i )
+    for( int i = num_bits; i > 0; --i )
       {
-      range >>= 1;
       symbol <<= 1;
-      if( ( ( code - range ) & 0x80000000 ) == 0 )
-        { code -= range; symbol |= 1; }
-      if( range <= 0xFFFFFF )
-        { range <<= 8; code = (code << 8) | read_byte(); }
+      if( range <= 0x00FFFFFF )
+        {
+        range <<= 7; code = (code << 8) | get_byte();
+        if( code >= range ) { code -= range; symbol |= 1; }
+        }
+      else
+        {
+        range >>= 1;
+        if( code >= range ) { code -= range; symbol |= 1; }
+        }
       }
     return symbol;
     }
 
   int decode_bit( Bit_model & bm )
     {
-    int symbol = 0;
+    normalize();
     const uint32_t bound = ( range >> bit_model_total_bits ) * bm.probability;
     if( code < bound )
       {
       range = bound;
       bm.probability += (bit_model_total - bm.probability) >> bit_model_move_bits;
+      return 0;
       }
     else
       {
       range -= bound;
       code -= bound;
       bm.probability -= bm.probability >> bit_model_move_bits;
-      symbol = 1;
+      return 1;
       }
-    if( range <= 0xFFFFFF )
-      { range <<= 8; code = (code << 8) | read_byte(); }
-    return symbol;
     }
 
   int decode_tree( Bit_model bm[], const int num_bits )
     {
     int model = 1;
     for( int i = num_bits; i > 0; --i )
-      model = ( model << 1 ) | decode_bit( bm[model-1] );
+      model = ( model << 1 ) | decode_bit( bm[model] );
     return model - (1 << num_bits);
     }
 
@@ -122,24 +139,29 @@ public:
     int symbol = 0;
     for( int i = 0; i < num_bits; ++i )
       {
-      model = ( model << 1 ) | decode_bit( bm[model-1] );
-      if( model & 1 ) symbol |= ( 1 << i );
+      const int bit = decode_bit( bm[model] );
+      model <<= 1;
+      if( bit ) { model |= 1; symbol |= (1 << i); }
       }
     return symbol;
     }
 
   int decode_matched( Bit_model bm[], const int match_byte )
     {
+    Bit_model *bm1 = bm + 0x100;
     int symbol = 1;
-    for( int i = 7; i >= 0; --i )
+    for( int i = 1; i <= 8; ++i )
       {
-      int match_bit = ( match_byte >> i ) & 1;
-      int bit = decode_bit( bm[(match_bit<<8)+symbol+0xFF] );
+      const int match_bit = ( match_byte << i ) & 0x100;
+      const int bit = decode_bit( bm1[match_bit+symbol] );
       symbol = ( symbol << 1 ) | bit;
-      if( match_bit != bit ) break;
+      if( ( match_bit && !bit ) || ( !match_bit && bit ) )
+        {
+        while( ++i <= 8 )
+          symbol = ( symbol << 1 ) | decode_bit( bm[symbol] );
+        break;
+        }
       }
-    while( symbol < 0x100 )
-      symbol = ( symbol << 1 ) | decode_bit( bm[symbol-1] );
     return symbol & 0xFF;
     }
   };
@@ -169,37 +191,32 @@ public:
 
 class Literal_decoder
   {
-  typedef Bit_model Bm_array[0x300];
-  Bm_array * const bm_literal;
+  Bit_model bm_literal[1<<literal_context_bits][0x300];
 
   int state( const int prev_byte ) const throw()
     { return ( prev_byte >> ( 8 - literal_context_bits ) ); }
 
 public:
-  Literal_decoder()
-    : bm_literal( new Bm_array[1<<literal_context_bits] ) {}
-
-  ~Literal_decoder() { delete[] bm_literal; }
-
-  uint8_t decode( Range_decoder & range_decoder, const int prev_byte )
+  uint8_t decode( Range_decoder & range_decoder, const uint8_t prev_byte )
     { return range_decoder.decode_tree( bm_literal[state(prev_byte)], 8 ); }
 
   uint8_t decode_matched( Range_decoder & range_decoder,
-                          const int prev_byte, const int match_byte )
+                          const uint8_t prev_byte, const uint8_t match_byte )
     { return range_decoder.decode_matched( bm_literal[state(prev_byte)], match_byte ); }
   };
 
 
 class LZ_decoder
   {
-  long long partial_file_pos;
+  long long partial_data_pos;
   const int format_version;
+  const int dictionary_size;
   const int buffer_size;
   uint8_t * const buffer;
   int pos;
-  int stream_pos;
-  const int _odes;
-  uint32_t _crc;
+  int stream_pos;		// first byte not yet written to file
+  uint32_t crc_;
+  const int odes_;		// output file descriptor
 
   Bit_model bm_match[State::states][pos_states];
   Bit_model bm_rep[State::states];
@@ -218,58 +235,60 @@ class LZ_decoder
 
   uint8_t get_byte( const int distance ) const throw()
     {
-    int newpos = pos - distance - 1;
-    if( newpos < 0 ) newpos += buffer_size;
-    return buffer[newpos];
+    int i = pos - distance - 1;
+    if( i < 0 ) i += buffer_size;
+    return buffer[i];
     }
 
   void put_byte( const uint8_t b )
     {
-    update_crc( _crc, b );
     buffer[pos] = b;
-    if( ++pos >= buffer_size ) flush();
+    if( ++pos >= buffer_size ) flush_data();
     }
 
-  bool copy_block( const int distance, int len )
+  void copy_block( const int distance, int len )
     {
-    if( distance < 0 || distance >= buffer_size ||
-        len <= 0 || len > max_match_len ) return false;
-    int newpos = pos - distance - 1;
-    if( newpos < 0 ) newpos += buffer_size;
-    for( ; len > 0 ; --len )
+    int i = pos - distance - 1;
+    if( i < 0 ) i += buffer_size;
+    if( len < buffer_size - std::max( pos, i ) && len <= std::abs( pos - i ) )
       {
-      update_crc( _crc, buffer[newpos] );
-      buffer[pos] = buffer[newpos];
-      if( ++pos >= buffer_size ) flush();
-      if( ++newpos >= buffer_size ) newpos = 0;
+      std::memcpy( buffer + pos, buffer + i, len );
+      pos += len;
+      }
+    else for( ; len > 0 ; --len )
+      {
+      buffer[pos] = buffer[i];
+      if( ++pos >= buffer_size ) flush_data();
+      if( ++i >= buffer_size ) i = 0;
       }
-    return true;
     }
 
-  void flush();
-  bool verify_trailer( const Pretty_print & pp );
+  void flush_data();
+  bool verify_trailer( const Pretty_print & pp ) const;
 
 public:
   LZ_decoder( const File_header & header, Input_buffer & ibuf, const int odes )
     :
-    partial_file_pos( 0 ),
+    partial_data_pos( 0 ),
     format_version( header.version ),
-    buffer_size( 1 << header.dictionary_bits ),
+    dictionary_size( header.dictionary_size() ),
+    buffer_size( std::max( 65536, dictionary_size ) ),
     buffer( new uint8_t[buffer_size] ),
     pos( 0 ),
     stream_pos( 0 ),
-    _odes( odes ),
-    _crc( 0xFFFFFFFF ),
+    crc_( 0xFFFFFFFF ),
+    odes_( odes ),
     range_decoder( sizeof header, ibuf ),
-    literal_decoder() {}
+    literal_decoder()
+    { buffer[buffer_size-1] = 0; }	// prev_byte of first_byte
 
   ~LZ_decoder() { delete[] buffer; }
 
-  uint32_t crc() const throw() { return _crc ^ 0xFFFFFFFF; }
-  int decode( const Pretty_print & pp );
+  uint32_t crc() const throw() { return crc_ ^ 0xFFFFFFFF; }
+  int decode_member( const Pretty_print & pp );
 
-  long long input_file_position() const throw()
-    { return range_decoder.file_position(); }
-  long long output_file_position() const throw()
-    { return partial_file_pos + pos; }
+  long long member_position() const throw()
+    { return range_decoder.member_position(); }
+  long long data_position() const throw()
+    { return partial_data_pos + pos; }
   };
