@@ -1,5 +1,5 @@
 /*  Lzip - A data compressor based on the LZMA algorithm
-    Copyright (C) 2008 Antonio Diaz Diaz.
+    Copyright (C) 2008, 2009, 2010 Antonio Diaz Diaz.
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -26,6 +26,7 @@
 #include <algorithm>
 #include <cerrno>
 #include <climits>
+#include <csignal>
 #include <cstdio>
 #include <cstdlib>
 #include <cstring>
@@ -33,7 +34,6 @@
 #include <vector>
 #include <fcntl.h>
 #include <stdint.h>
-#include <signal.h>
 #include <unistd.h>
 #include <utime.h>
 #include <sys/stat.h>
@@ -60,7 +60,7 @@ namespace {
 const char * invocation_name = 0;
 const char * const Program_name    = "Lzip";
 const char * const program_name    = "lzip";
-const char * const program_year    = "2008";
+const char * const program_year    = "2010";
 
 struct { const char * from; const char * to; } const known_extensions[] = {
   { ".lz",  ""     },
@@ -69,13 +69,12 @@ struct { const char * from; const char * to; } const known_extensions[] = {
 
 struct lzma_options
   {
-  int dictionary_bits;		// 12..29
+  int dictionary_size;		// 4KiB..512MiB
   int match_len_limit;		// 5..273
   };
 
 enum Mode { m_compress = 0, m_decompress, m_test };
 
-std::string input_filename;
 std::string output_filename;
 int outhandle = -1;
 bool delete_output_on_interrupt = false;
@@ -85,25 +84,30 @@ void show_help() throw()
   {
   std::printf( "%s - A data compressor based on the LZMA algorithm.\n", Program_name );
   std::printf( "\nUsage: %s [options] [files]\n", invocation_name );
-  std::printf( "Options:\n" );
+  std::printf( "\nOptions:\n" );
   std::printf( "  -h, --help                 display this help and exit\n" );
   std::printf( "  -V, --version              output version information and exit\n" );
+  std::printf( "  -b, --member-size=<n>      set member size limit in bytes\n" );
   std::printf( "  -c, --stdout               send output to standard output\n" );
-  std::printf( "  -d, --decompress           force decompression\n" );
+  std::printf( "  -d, --decompress           decompress\n" );
   std::printf( "  -f, --force                overwrite existing output files\n" );
   std::printf( "  -k, --keep                 keep (don't delete) input files\n" );
-  std::printf( "  -m, --match-length=<n>     set match length limit in bytes [64]\n" );
+  std::printf( "  -m, --match-length=<n>     set match length limit in bytes [80]\n" );
+  std::printf( "  -o, --output=<file>        if reading stdin, place the output into <file>\n" );
   std::printf( "  -q, --quiet                suppress all messages\n" );
   std::printf( "  -s, --dictionary-size=<n>  set dictionary size limit in bytes [8MiB]\n" );
+  std::printf( "  -S, --volume-size=<n>      set volume size limit in bytes\n" );
   std::printf( "  -t, --test                 test compressed file integrity\n" );
   std::printf( "  -v, --verbose              be verbose (a 2nd -v gives more)\n" );
-  std::printf( "  -z, --compress             force compression\n" );
   std::printf( "  -1 .. -9                   set compression level [default 6]\n" );
   std::printf( "      --fast                 alias for -1\n" );
   std::printf( "      --best                 alias for -9\n" );
-  std::printf( "If no file names are given, lzip compresses or decompresses\n" );
+  std::printf( "If no file names are given, %s compresses or decompresses\n", program_name );
   std::printf( "from standard input to standard output.\n" );
-  std::printf( "\nReport bugs to lzip-bug@nongnu.org\n");
+  std::printf( "Numbers may be followed by a multiplier: k = kB = 10^3 = 1000,\n" );
+  std::printf( "Ki = KiB = 2^10 = 1024, M = 10^6, Mi = 2^20, G = 10^9, Gi = 2^30, etc...\n" );
+  std::printf( "\nReport bugs to lzip-bug@nongnu.org\n" );
+  std::printf( "Lzip home page: http://www.nongnu.org/lzip/lzip.html\n" );
   }
 
 
@@ -117,7 +121,7 @@ void show_version() throw()
   }
 
 
-const char * format_num( long long num, long long max = 1023,
+const char * format_num( long long num, long long limit = 9999,
                          const int set_prefix = 0 ) throw()
   {
   const char * const si_prefix[8] =
@@ -131,18 +135,19 @@ const char * format_num( long long num, long long max = 1023,
   const int factor = ( si ) ? 1000 : 1024;
   const char * const *prefix = ( si ) ? si_prefix : binary_prefix;
   const char *p = "";
-  max = std::max( 999LL, std::min( 999999LL, max ) );
+  limit = std::max( 999LL, std::min( 999999LL, limit ) );
 
-  for( int i = 0; i < 8 && llabs( num ) > llabs( max ); ++i )
+  for( int i = 0; i < 8 && ( llabs( num ) > limit ||
+       ( llabs( num ) >= factor && num % factor == 0 ) ); ++i )
     { num /= factor; p = prefix[i]; }
   snprintf( buf, sizeof buf, "%lld %s", num, p );
   return buf;
   }
 
 
-long long getnum( const char * ptr, const int bs,
-                  const long long min = LLONG_MIN + 1,
-                  const long long max = LLONG_MAX ) throw()
+long long getnum( const char * ptr, const int bs = 0,
+                  const long long llimit = LLONG_MIN + 1,
+                  const long long ulimit = LLONG_MAX ) throw()
   {
   errno = 0;
   char *tail;
@@ -175,7 +180,7 @@ long long getnum( const char * ptr, const int bs,
                 break;
       case 'k': if( factor == 1000 ) exponent = 1; else bad_multiplier = true;
                 break;
-      default: bad_multiplier = true;
+      default : bad_multiplier = true;
       }
     if( bad_multiplier )
       {
@@ -188,7 +193,7 @@ long long getnum( const char * ptr, const int bs,
       else { errno = ERANGE; break; }
       }
     }
-  if( !errno && ( result < min || result > max ) ) errno = ERANGE;
+  if( !errno && ( result < llimit || result > ulimit ) ) errno = ERANGE;
   if( errno )
     {
     show_error( "numerical argument out of limits" );
@@ -198,253 +203,123 @@ long long getnum( const char * ptr, const int bs,
   }
 
 
-int real_bits( const int value ) throw()
-  {
-  int bits = 0;
-  for( int i = 1, mask = 1; mask > 0; ++i, mask <<= 1 )
-    if( value & mask ) bits = i;
-  return bits;
-  }
-
-
-int get_dict_bits( const char * arg ) throw()
+int get_dict_size( const char * arg ) throw()
   {
   char *tail;
   int bits = std::strtol( arg, &tail, 0 );
   if( bits >= min_dictionary_bits && bits <= max_dictionary_bits && *tail == 0 )
-    return bits;
-  const int min_size = 1 << min_dictionary_bits;
-  const int max_size = 1 << max_dictionary_bits;
-  const int size = getnum( arg, 0, ( min_size / 2 ) + 1, max_size );
-  bits = real_bits( size - 1 );
-  if( size > ( 1 << bits ) ) ++bits;
-  return bits;
+    return ( 1 << bits );
+  return getnum( arg, 0, min_dictionary_size, max_dictionary_size );
   }
 
 
-int compress( const int ides, const int odes, lzma_options encoder_options,
-              const Pretty_print & pp ) throw()
-  {
-  if( verbosity >= 1 ) pp();
-  if( encoder_options.dictionary_bits < min_dictionary_bits ||
-      encoder_options.dictionary_bits > max_dictionary_bits ||
-      encoder_options.match_len_limit < 5 ||
-      encoder_options.match_len_limit > max_match_len )
-    internal_error( "invalid argument to encoder" );
-
-  File_header header;
-  header.set_magic();
-  header.dictionary_bits = encoder_options.dictionary_bits;
-
-  try {
-    LZ_encoder encoder( header, ides, odes, encoder_options.match_len_limit );
-
-    if( !encoder.encode() ) { pp( "encoder error" ); return 2; }
-
-    if( verbosity >= 1 )
-      {
-      long long in_size = encoder.input_file_position();
-      long long out_size = encoder.output_file_position();
-
-      if( in_size <= 0 || out_size <= 0 )
-        std::fprintf( stderr, "no data compressed.\n" );
-      else
-        std::fprintf( stderr, "%6.3f:1, %6.3f bits/byte, "
-                              "%5.2f%% saved, %lld in, %lld out.\n",
-                      (double)in_size / out_size,
-                      ( 8.0 * out_size ) / in_size,
-                      100.0 * ( 1.0 - ( (double)out_size / in_size ) ),
-                      in_size, out_size );
-      }
-    }
-  catch( std::bad_alloc )
-    {
-    pp( "not enough memory. Try a smaller dictionary size" );
-    return 1;
-    }
-  catch( Error e ) { pp(); show_error( e.s, errno ); return 1; }
-  return 0;
-  }
-
-
-int decompress( const int ides, const int odes, const Pretty_print & pp,
-                const bool testing ) throw()
-  {
-  Input_buffer ibuf( ides );
-  for( bool first_pass = true; ; first_pass = false, pp.reset() )
-    {
-    File_header header;
-    for( unsigned int i = 0; i < sizeof header; ++i )
-      ((uint8_t *)&header)[i] = ibuf.read_byte();
-    if( ibuf.finished() )
-      {
-      if( first_pass ) { pp( "error reading file header" ); return 1; }
-      else break;
-      }
-    if( !header.verify_magic() )
-      {
-      if( !first_pass ) break;
-      if( verbosity >= 0 )
-        { pp();
-          std::fprintf( stderr, "bad magic number (file not created by %s).\n",
-                        program_name ); }
-      return 2;
-      }
-    if( !header.verify_version() )
-      {
-      if( verbosity >= 0 )
-        { pp();
-          std::fprintf( stderr, "version %d, file format not supported, newer %s needed.\n",
-                        header.version, program_name ); }
-      return 2;
-      }
-    if( header.dictionary_bits < min_dictionary_bits ||
-        header.dictionary_bits > max_dictionary_bits )
-      { pp( "invalid dictionary size in file header" ); return 2; }
-
-    try {
-      if( verbosity >= 1 )
-        {
-        pp();
-        if( verbosity >= 2 )
-          std::fprintf( stderr, "version %d, dictionary size %6sB.  ",
-                        header.version,
-                        format_num( 1 << header.dictionary_bits ) );
-        }
-      LZ_decoder decoder( header, ibuf, odes );
-
-      const int result = decoder.decode( pp );
-      if( result != 0 )
-        {
-        if( verbosity >= 0 && result <= 2 )
-          {
-          pp();
-          if( result == 1 )
-            std::fprintf( stderr, "decoder error at pos %lld\n",
-                          decoder.input_file_position() );
-          if( result == 2 )
-            std::fprintf( stderr, "file ends unexpectedly at pos %lld\n",
-                          decoder.input_file_position() );
-          }
-        return 2;
-        }
-      if( verbosity >= 1 )
-        { if( testing ) std::fprintf( stderr, "ok\n" );
-          else std::fprintf( stderr, "done\n" ); }
-      }
-    catch( std::bad_alloc )
-      {
-      pp( "not enough memory. Find a machine with more memory" );
-      return 1;
-      }
-    catch( Error e ) { pp(); show_error( e.s, errno ); return 1; }
-    }
-  return 0;
-  }
-
-
-int extension_index() throw()
+int extension_index( const std::string & name ) throw()
   {
   for( int i = 0; known_extensions[i].from; ++i )
     {
     const std::string ext( known_extensions[i].from );
-    if( input_filename.size() > ext.size() &&
-        input_filename.compare( input_filename.size() - ext.size(),
-                                ext.size(), ext ) == 0 )
+    if( name.size() > ext.size() &&
+        name.compare( name.size() - ext.size(), ext.size(), ext ) == 0 )
       return i;
     }
   return -1;
   }
 
 
-std::string replace_extension( const int i ) throw()
-  {
-  if( i >= 0 )
-    {
-    const std::string from = known_extensions[i].from;
-    if( input_filename.size() > from.size() )
-      {
-      const int len = input_filename.size() - from.size();
-      std::string out( input_filename, 0, len );
-      out += known_extensions[i].to;
-      return out;
-      }
-    }
-  std::string out( input_filename ); out += ".out";
-  if( verbosity >= 0 )
-    std::fprintf( stderr, "%s: can't guess original name for `%s' -- using `%s'.\n",
-                  program_name, input_filename.c_str(), out.c_str() );
-  return out;
-  }
-
-
-int open_instream( struct stat * in_statsp, const Mode program_mode,
-                   const int eindex, const bool force ) throw()
+int open_instream( const std::string & name, struct stat * in_statsp,
+                   const Mode program_mode, const int eindex,
+                   const bool force, const bool to_stdout ) throw()
   {
+  int inhandle = -1;
   if( program_mode == m_compress && !force && eindex >= 0 )
     {
     if( verbosity >= 0 )
       std::fprintf( stderr, "%s: input file `%s' already has `%s' suffix.\n",
-                    program_name, input_filename.c_str(),
+                    program_name, name.c_str(),
                     known_extensions[eindex].from );
-    return -1;
-    }
-
-  int ides = open( input_filename.c_str(), O_RDONLY );
-  if( ides < 0 )
-    {
-    if( verbosity >= 0 )
-      std::fprintf( stderr, "%s: Can't open input file `%s': %s.\n",
-                    program_name, input_filename.c_str(), strerror( errno ) );
     }
   else
     {
-    const int i = fstat( ides, in_statsp );
-    if( i < 0 || !S_ISREG( in_statsp->st_mode ) )
+    inhandle = open( name.c_str(), O_RDONLY );
+    if( inhandle < 0 )
       {
       if( verbosity >= 0 )
-        std::fprintf( stderr, "%s: input file `%s' is not a regular file.\n",
-                      program_name, input_filename.c_str() );
-      close( ides );
-      ides = -1;
+        std::fprintf( stderr, "%s: Can't open input file `%s': %s.\n",
+                      program_name, name.c_str(), std::strerror( errno ) );
+      }
+    else
+      {
+      const int i = fstat( inhandle, in_statsp );
+      const mode_t & mode = in_statsp->st_mode;
+      if( i < 0 || !( S_ISREG( mode ) || ( to_stdout &&
+                      ( S_ISFIFO( mode ) || S_ISSOCK( mode ) ||
+                        S_ISBLK( mode ) || S_ISCHR( mode ) ) ) ) )
+        {
+        if( verbosity >= 0 )
+          std::fprintf( stderr, "%s: input file `%s' is not a regular file%s.\n",
+                        program_name, name.c_str(),
+                        to_stdout ? "" : " and `--stdout' was not specified" );
+        close( inhandle );
+        inhandle = -1;
+        }
       }
     }
-  return ides;
+  return inhandle;
   }
 
 
-int open_outstream( const Mode program_mode, const int eindex,
-                    const bool force ) throw()
+void set_c_outname( const std::string & name, const bool multifile ) throw()
   {
-  if( program_mode == m_test )
-    output_filename = "/dev/null";
-  else if( program_mode == m_decompress )
-    output_filename = replace_extension( eindex );
-  else
-    output_filename = input_filename + known_extensions[0].from;
+  output_filename = name;
+  if( multifile ) output_filename += "00001";
+  output_filename += known_extensions[0].from;
+  }
 
-  int odes;
-  if( force || ( program_mode == m_test ) )
-    odes = open( output_filename.c_str(), O_CREAT | O_TRUNC | O_WRONLY, S_IRUSR | S_IWUSR );
-  else odes = open( output_filename.c_str(), O_CREAT | O_EXCL | O_WRONLY, S_IRUSR | S_IWUSR );
-  if( odes < 0 )
+
+void set_d_outname( const std::string & name, const int i ) throw()
+  {
+  if( i >= 0 )
     {
-    if( errno == EEXIST ) odes = -2; else odes = -1;
+    const std::string from( known_extensions[i].from );
+    if( name.size() > from.size() )
+      {
+      output_filename.assign( name, 0, name.size() - from.size() );
+      output_filename += known_extensions[i].to;
+      return;
+      }
+    }
+  output_filename = name; output_filename += ".out";
+  if( verbosity >= 0 )
+    std::fprintf( stderr, "%s: can't guess original name for `%s' -- using `%s'.\n",
+                  program_name, name.c_str(), output_filename.c_str() );
+  }
+
+
+bool open_outstream( const bool force ) throw()
+  {
+  if( force )
+    outhandle = open( output_filename.c_str(), O_CREAT | O_TRUNC | O_WRONLY,
+                      S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH );
+  else outhandle = open( output_filename.c_str(), O_CREAT | O_EXCL | O_WRONLY,
+                         S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH );
+  if( outhandle < 0 )
+    {
+    if( errno == EEXIST ) outhandle = -2; else outhandle = -1;
     if( verbosity >= 0 )
       {
-      if( odes == -2 )
+      if( outhandle == -2 )
         std::fprintf( stderr, "%s: Output file %s already exists, skipping.\n",
                       program_name, output_filename.c_str() );
       else
         std::fprintf( stderr, "%s: Can't create output file `%s': %s.\n",
-                      program_name, output_filename.c_str(), strerror( errno ) );
+                      program_name, output_filename.c_str(), std::strerror( errno ) );
       }
     }
-  return odes;
+  return ( outhandle >= 0 );
   }
 
 
-bool check_tty( const Mode program_mode, const int inhandle ) throw()
+bool check_tty( const int inhandle, const Mode program_mode ) throw()
   {
   if( program_mode == m_compress && isatty( outhandle ) )
     {
@@ -476,18 +351,206 @@ void cleanup_and_fail( const int retval ) throw()
   }
 
 
-void signal_handler( const int ) throw()
+     // Set permissions, owner and times.
+void close_and_set_permissions( const struct stat * const in_statsp )
+  {
+  bool error = false;
+  if( in_statsp )
+    {
+    if( fchmod( outhandle, in_statsp->st_mode ) != 0 ) error = true;
+    else (void)fchown( outhandle, in_statsp->st_uid, in_statsp->st_gid );
+    // fchown will in many cases return with EPERM, which can be safely ignored.
+    }
+  if( close( outhandle ) == 0 ) outhandle = -1;
+  else cleanup_and_fail( 1 );
+  delete_output_on_interrupt = false;
+  if( !in_statsp ) return;
+  if( !error )
+    {
+    struct utimbuf t;
+    t.actime = in_statsp->st_atime;
+    t.modtime = in_statsp->st_mtime;
+    if( utime( output_filename.c_str(), &t ) != 0 ) error = true;
+    }
+  if( error )
+    {
+    show_error( "I can't change output file attributes." );
+    cleanup_and_fail( 1 );
+    }
+  }
+
+
+bool next_filename()
+  {
+  const unsigned int len = std::strlen( known_extensions[0].from );
+  if( output_filename.size() >= len + 5 )		// "*00001.lz"
+    for( int i = output_filename.size() - len - 1, j = 0; j < 5; --i, ++j )
+      {
+      if( output_filename[i] < '9' ) { ++output_filename[i]; return true; }
+      else output_filename[i] = '0';
+      }
+  return false;
+  }
+
+
+int compress( const long long member_size, const long long volume_size,
+              const lzma_options & encoder_options, const int inhandle,
+              const Pretty_print & pp, const struct stat * const in_statsp )
+  {
+  if( verbosity >= 1 ) pp();
+  File_header header;
+  header.set_magic();
+  if( !header.dictionary_size( encoder_options.dictionary_size ) ||
+      encoder_options.match_len_limit < min_match_len_limit ||
+      encoder_options.match_len_limit > max_match_len )
+    internal_error( "invalid argument to encoder" );
+
+  try {
+    Matchfinder matchfinder( header.dictionary_size(),
+                             encoder_options.match_len_limit, inhandle );
+    header.dictionary_size( matchfinder.dictionary_size() );
+
+    long long in_size = 0, out_size = 0, partial_volume_size = 0;
+    while( true )		// encode one member per iteration
+      {
+      LZ_encoder encoder( matchfinder, header, outhandle );
+      const long long size =
+        std::min( member_size, volume_size - partial_volume_size );
+      if( !encoder.encode_member( size ) )
+        { pp(); show_error( "read error", errno ); return 1; }
+      in_size += matchfinder.data_position();
+      out_size += encoder.member_position();
+      if( matchfinder.finished() ) break;
+      partial_volume_size += encoder.member_position();
+      if( partial_volume_size >= volume_size - min_dictionary_size )
+        {
+        partial_volume_size = 0;
+        if( delete_output_on_interrupt )
+          {
+          close_and_set_permissions( in_statsp );
+          if( !next_filename() )
+            { pp(); show_error( "too many volume files" ); return 1; }
+          if( !open_outstream( true ) ) return 1;
+          delete_output_on_interrupt = true;
+          }
+        }
+      if( !matchfinder.reset() )
+        { pp(); show_error( "read error", errno ); return 1; }
+      }
+
+    if( verbosity >= 1 )
+      {
+      if( in_size <= 0 || out_size <= 0 )
+        std::fprintf( stderr, "no data compressed.\n" );
+      else
+        std::fprintf( stderr, "%6.3f:1, %6.3f bits/byte, "
+                              "%5.2f%% saved, %lld in, %lld out.\n",
+                      (double)in_size / out_size,
+                      ( 8.0 * out_size ) / in_size,
+                      100.0 * ( 1.0 - ( (double)out_size / in_size ) ),
+                      in_size, out_size );
+      }
+    }
+  catch( std::bad_alloc )
+    {
+    pp( "not enough memory. Try a smaller dictionary size" );
+    return 1;
+    }
+  catch( Error e ) { pp(); show_error( e.s, errno ); return 1; }
+  return 0;
+  }
+
+
+int decompress( const int inhandle, const Pretty_print & pp,
+                const bool testing )
+  {
+  try {
+    Input_buffer ibuf( inhandle );
+    long long partial_file_pos = 0;
+    for( bool first_member = true; ; first_member = false, pp.reset() )
+      {
+      File_header header;
+      for( unsigned int i = 0; i < sizeof header; ++i )
+        ((uint8_t *)&header)[i] = ibuf.get_byte();
+      if( ibuf.finished() )			// End Of File
+        {
+        if( !first_member ) break;
+        pp( "error reading member header" ); return 1;
+        }
+      if( !header.verify_magic() )
+        {
+        if( !first_member ) break;		// trailing garbage
+        if( verbosity >= 0 )
+          { pp();
+            std::fprintf( stderr, "bad magic number (file not created by %s).\n",
+                          program_name ); }
+        return 2;
+        }
+      if( !header.verify_version() )
+        {
+        if( verbosity >= 0 )
+          { pp();
+            std::fprintf( stderr, "version %d member format not supported, newer %s needed.\n",
+                          header.version, program_name ); }
+        return 2;
+        }
+      if( header.dictionary_size() < min_dictionary_size ||
+          header.dictionary_size() > max_dictionary_size )
+        { pp( "invalid dictionary size in member header" ); return 2; }
+
+      if( verbosity >= 1 )
+        {
+        pp();
+        if( verbosity >= 2 )
+          std::fprintf( stderr, "version %d, dictionary size %7sB.  ",
+                        header.version,
+                        format_num( header.dictionary_size() ) );
+        }
+      LZ_decoder decoder( header, ibuf, outhandle );
+
+      const int result = decoder.decode_member( pp );
+      partial_file_pos += decoder.member_position();
+      if( result != 0 )
+        {
+        if( verbosity >= 0 && result <= 2 )
+          {
+          pp();
+          if( result == 2 )
+            std::fprintf( stderr, "file ends unexpectedly at pos %lld\n",
+                          partial_file_pos );
+          else
+            std::fprintf( stderr, "decoder error at pos %lld\n",
+                          partial_file_pos );
+          }
+        return 2;
+        }
+      if( verbosity >= 1 )
+        { if( testing ) std::fprintf( stderr, "ok\n" );
+          else std::fprintf( stderr, "done\n" ); }
+      }
+    }
+  catch( std::bad_alloc )
+    {
+    pp( "not enough memory. Find a machine with more memory" );
+    return 1;
+    }
+  catch( Error e ) { pp(); show_error( e.s, errno ); return 1; }
+  return 0;
+  }
+
+
+extern "C" void signal_handler( int ) throw()
   {
   show_error( "Control-C or similar caught, quitting." );
-  cleanup_and_fail( 0 );
+  cleanup_and_fail( 1 );
   }
 
 
 void set_signals() throw()
   {
-  signal( SIGTERM, signal_handler );
   signal( SIGHUP, signal_handler );
   signal( SIGINT, signal_handler );
+  signal( SIGTERM, signal_handler );
   }
 
 } // end namespace
@@ -503,8 +566,8 @@ void Pretty_print::operator()( const char * const msg ) const throw()
     if( first_post )
       {
       first_post = false;
-      std::fprintf( stderr, "  %s: ", _name.c_str() );
-      for( unsigned int i = 0; i < longest_name - _name.size(); ++i )
+      std::fprintf( stderr, "  %s: ", name_.c_str() );
+      for( unsigned int i = 0; i < longest_name - name_.size(); ++i )
         std::fprintf( stderr, " " );
       if( !msg ) std::fflush( stderr );
       }
@@ -520,7 +583,7 @@ void show_error( const char * msg, const int errcode, const bool help ) throw()
     if( msg && msg[0] != 0 )
       {
       std::fprintf( stderr, "%s: %s", program_name, msg );
-      if( errcode > 0 ) std::fprintf( stderr, ": %s", strerror( errcode ) );
+      if( errcode > 0 ) std::fprintf( stderr, ": %s", std::strerror( errcode ) );
       std::fprintf( stderr, "\n" );
       }
     if( help && invocation_name && invocation_name[0] != 0 )
@@ -529,7 +592,7 @@ void show_error( const char * msg, const int errcode, const bool help ) throw()
   }
 
 
-void internal_error( const char * msg ) throw()
+void internal_error( const char * msg )
   {
   std::string s( "internal error: " ); s += msg;
   show_error( s.c_str() );
@@ -580,23 +643,28 @@ int main( const int argc, const char * argv[] )
   // to the corresponding LZMA compression modes.
   const lzma_options option_mapping[] =
     {
-    { 22,  10 },		// -1
-    { 22,  12 },		// -2
-    { 22,  16 },		// -3
-    { 22,  32 },		// -4
-    { 22,  64 },		// -5
-    { 23,  64 },		// -6
-    { 24,  64 },		// -7
-    { 24, 128 },		// -8
-    { 25, 273 } };		// -9
+    { 1 << 20,  10 },		// -1
+    { 1 << 20,  12 },		// -2
+    { 1 << 20,  17 },		// -3
+    { 1 << 21,  26 },		// -4
+    { 1 << 22,  44 },		// -5
+    { 1 << 23,  80 },		// -6
+    { 1 << 24, 108 },		// -7
+    { 1 << 24, 163 },		// -8
+    { 1 << 25, 273 } };		// -9
   lzma_options encoder_options = option_mapping[5];	// default = "-6"
-  Mode program_mode = m_compress;
+  long long member_size = LLONG_MAX;
+  long long volume_size = LLONG_MAX;
   int inhandle = -1;
+  Mode program_mode = m_compress;
   bool force = false;
   bool keep_input_files = false;
   bool to_stdout = false;
+  std::string input_filename;
+  std::string default_output_filename;
   std::vector< std::string > filenames;
   invocation_name = argv[0];
+//  std::setvbuf( stderr, 0, _IONBF, 0 );
 
   const Arg_parser::Option options[] =
     {
@@ -609,18 +677,20 @@ int main( const int argc, const char * argv[] )
     { '7',  0,                Arg_parser::no  },
     { '8',  0,                Arg_parser::no  },
     { '9', "best",            Arg_parser::no  },
+    { 'b', "member-size",     Arg_parser::yes },
     { 'c', "stdout",          Arg_parser::no  },
     { 'd', "decompress",      Arg_parser::no  },
     { 'f', "force",           Arg_parser::no  },
     { 'h', "help",            Arg_parser::no  },
     { 'k', "keep",            Arg_parser::no  },
     { 'm', "match-length",    Arg_parser::yes },
+    { 'o', "output",          Arg_parser::yes },
     { 'q', "quiet",           Arg_parser::no  },
     { 's', "dictionary-size", Arg_parser::yes },
+    { 'S', "volume-size",     Arg_parser::yes },
     { 't', "test",            Arg_parser::no  },
     { 'v', "verbose",         Arg_parser::no  },
     { 'V', "version",         Arg_parser::no  },
-    { 'z', "compress",        Arg_parser::no  },
     {  0 ,  0,                Arg_parser::no  } };
 
   Arg_parser parser( argc, argv, options );
@@ -639,20 +709,22 @@ int main( const int argc, const char * argv[] )
       case '4': case '5': case '6':
       case '7': case '8': case '9':
                 encoder_options = option_mapping[code-'1']; break;
+      case 'b': member_size = getnum( arg, 0, 100000, LLONG_MAX / 2 ); break;
       case 'c': to_stdout = true; break;
       case 'd': program_mode = m_decompress; break;
       case 'f': force = true; break;
       case 'h': show_help(); return 0;
       case 'k': keep_input_files = true; break;
       case 'm': encoder_options.match_len_limit =
-                getnum( arg, 0, 5, max_match_len ); break;
+                getnum( arg, 0, min_match_len_limit, max_match_len ); break;
+      case 'o': default_output_filename = arg; break;
       case 'q': verbosity = -1; break;
-      case 's': encoder_options.dictionary_bits = get_dict_bits( arg );
+      case 's': encoder_options.dictionary_size = get_dict_size( arg );
                 break;
+      case 'S': volume_size = getnum( arg, 0, 100000, LLONG_MAX / 2 ); break;
       case 't': program_mode = m_test; break;
       case 'v': if( verbosity < 4 ) ++verbosity; break;
       case 'V': show_version(); return 0;
-      case 'z': program_mode = m_compress; break;
       default : internal_error( "uncaught option" );
       }
     }
@@ -669,9 +741,11 @@ int main( const int argc, const char * argv[] )
 
   Pretty_print pp( filenames );
   if( program_mode == m_test )
+    outhandle = -1;
+  else if( program_mode == m_compress )
     {
-    outhandle = open_outstream( program_mode, 0, force );
-    if( outhandle < 0 ) return 1;
+    dis_slots.init();
+    prob_prices.init();
     }
 
   int retval = 0;
@@ -684,21 +758,40 @@ int main( const int argc, const char * argv[] )
       {
       input_filename.clear();
       inhandle = STDIN_FILENO;
-      if( program_mode != m_test ) outhandle = STDOUT_FILENO;
+      if( program_mode != m_test )
+        {
+        if( to_stdout || !default_output_filename.size() )
+          outhandle = STDOUT_FILENO;
+        else
+          {
+          if( program_mode == m_compress )
+            set_c_outname( default_output_filename, volume_size != LLONG_MAX );
+          else output_filename = default_output_filename;
+          if( !open_outstream( force ) )
+            {
+            if( outhandle == -1 && retval < 1 ) retval = 1;
+            close( inhandle ); inhandle = -1;
+            continue;
+            }
+          }
+        }
       }
     else
       {
       input_filename = filenames[i];
-      const int eindex = extension_index();
-      inhandle = open_instream( &in_stats, program_mode, eindex, force );
-      if( inhandle < 0 ) continue;
+      const int eindex = extension_index( input_filename );
+      inhandle = open_instream( input_filename, &in_stats, program_mode,
+                                eindex, force, to_stdout );
+      if( inhandle < 0 ) { if( retval < 1 ) retval = 1; continue; }
       if( program_mode != m_test )
         {
         if( to_stdout ) outhandle = STDOUT_FILENO;
         else
           {
-          outhandle = open_outstream( program_mode, eindex, force );
-          if( outhandle < 0 )
+          if( program_mode == m_compress )
+            set_c_outname( input_filename, volume_size != LLONG_MAX );
+          else set_d_outname( input_filename, eindex );
+          if( !open_outstream( force ) )
             {
             if( outhandle == -1 && retval < 1 ) retval = 1;
             close( inhandle ); inhandle = -1;
@@ -708,41 +801,23 @@ int main( const int argc, const char * argv[] )
         }
       }
 
-    if( !check_tty( program_mode, inhandle ) ) return 1;
+    if( !check_tty( inhandle, program_mode ) ) return 1;
 
-    if( input_filename.size() && !to_stdout && program_mode != m_test )
+    if( output_filename.size() && !to_stdout && program_mode != m_test )
       delete_output_on_interrupt = true;
+    const struct stat * const in_statsp = input_filename.size() ? &in_stats : 0;
     pp.set_name( input_filename );
     int tmp = 0;
     if( program_mode == m_compress )
-      tmp = compress( inhandle, outhandle, encoder_options, pp );
+      tmp = compress( member_size, volume_size, encoder_options, inhandle,
+                      pp, in_statsp );
     else
-      tmp = decompress( inhandle, outhandle, pp, program_mode == m_test );
+      tmp = decompress( inhandle, pp, program_mode == m_test );
     if( tmp > retval ) retval = tmp;
     if( tmp && program_mode != m_test ) cleanup_and_fail( retval );
 
-    // Set permissions, owner and times.
-    if( input_filename.size() && !to_stdout && program_mode != m_test )
-      {
-      tmp = 0;
-      if( fchmod( outhandle, in_stats.st_mode ) != 0 ) tmp = 1;
-      if( !tmp ) fchown( outhandle, in_stats.st_uid, in_stats.st_gid );
-      if( close( outhandle ) != 0 ) cleanup_and_fail( 1 );
-      delete_output_on_interrupt = false;
-      if( !tmp )
-	{
-        struct utimbuf t;
-        t.actime = in_stats.st_atime;
-        t.modtime = in_stats.st_mtime;
-        tmp = utime( output_filename.c_str(), &t );
-        }
-      if( tmp )
-	{
-        if( tmp > retval ) retval = tmp;
-        show_error( "I can't change output file attributes." );
-        cleanup_and_fail( retval );
-	}
-      }
+    if( delete_output_on_interrupt )
+      close_and_set_permissions( in_statsp );
     if( input_filename.size() )
       {
       close( inhandle ); inhandle = -1;
@@ -750,5 +825,12 @@ int main( const int argc, const char * argv[] )
         std::remove( input_filename.c_str() );
       }
     }
+  if( outhandle >= 0 && close( outhandle ) != 0 )
+    {
+    if( verbosity >= 0 )
+      std::fprintf( stderr, "%s: Can't close stdout: %s.\n",
+                    program_name, std::strerror( errno ) );
+    if( retval < 1 ) retval = 1;
+    }
   return retval;
   }
